package core

import (
	"bufio"
	"io"
	"net"
	"time"

	"github.com/grufwub/go-errors"
)

// deadlineConn wraps net.Conn to set the read / write deadlines on each access
type deadlineConn struct {
	conn net.Conn
}

// Read wraps the underlying net.Conn read function, setting read deadline on each access
func (c *deadlineConn) Read(b []byte) (int, error) {
	c.conn.SetReadDeadline(time.Now().Add(connReadDeadline))
	return c.conn.Read(b)
}

// Read wraps the underlying net.Conn write function, setting write deadline on each access
func (c *deadlineConn) Write(b []byte) (int, error) {
	c.conn.SetWriteDeadline(time.Now().Add(connWriteDeadline))
	return c.conn.Write(b)
}

// Close directly wraps underlying net.Conn close function
func (c *deadlineConn) Close() error {
	return c.conn.Close()
}

// Conn wraps a DeadlineConn with a buffer
type conn struct {
	b  []byte
	br *bufio.Reader
	bw *bufio.Writer
	c  net.Conn
}

// wrapConn wraps a net.Conn in deadlineConn, then within conn and returns the result
func wrapConn(c net.Conn) *conn {
	deadlineConn := &deadlineConn{c}
	return &conn{
		b:  connRequestBufferPool.Get(),
		br: connBufferedReaderPool.Get(deadlineConn),
		bw: connBufferedWriterPool.Get(deadlineConn),
		c:  c,
	}
}

// Conn returns the underlying net.Conn
func (c *conn) Conn() net.Conn {
	return c.c
}

// ReadLine reads a single line and returns the result, or nil and error
func (c *conn) ReadLine() ([]byte, *errors.Error) {
	// Perform single read into buffer slice
	count, err := c.br.Read(c.b)
	if err != nil {
		return nil, errors.WrapError(ConnReadErr, err)
	}

	// Iterate up to '\r\n'
	end := -1
	for i := 0; i < count; i++ {
		if c.b[i] == '\r' {
			if i < count-1 && c.b[i+1] == '\n' {
				end = i
				break
			}
		}
	}

	// If end never reached return error,
	// else return slice up to the end
	if end == -1 {
		return nil, errors.NewError(InvalidRequestErr)
	}
	return c.b[:end], nil
}

// WriteBytes writes a byte slice to the buffer and returns error status
func (c *conn) Write(b []byte) *errors.Error {
	_, err := c.bw.Write(b)
	return errors.WrapIfError(ConnWriteErr, err)
}

// ReadFrom writes to the buffer from a reader and returns error status
func (c *conn) ReadFrom(r io.Reader) *errors.Error {
	// Since this buffer wraps deadlineConn, which DOES NOT have
	// a ReadFrom method implemented, it will force the buffer to
	// use it's own internal byte buffer along with the deadlineConn's
	// Write implementation (forcing the deadline to be regularly updated)
	_, err := c.bw.ReadFrom(r)
	return errors.WrapIfError(ConnWriteErr, err)
}

// Writer returns the underlying buffer wrapped conn writer
func (c *conn) Writer() io.Writer {
	return c.bw
}

// Close flushes the underlying buffer, closes the conn then puts
// the sync.Pool conn buffers back
func (c *conn) Close() *errors.Error {
	// Flush + close
	c.bw.Flush()
	err := c.c.Close()

	// Put buffers back
	connBufferedReaderPool.Put(c.br)
	connBufferedWriterPool.Put(c.bw)
	connRequestBufferPool.Put(c.b)

	// Return error (if exists)
	return errors.WrapIfError(ConnCloseErr, err)
}
