package core

import (
	"path"
	"strings"
)

// BuildPath is the global Path builder function for a supplied raw (relative) path
var BuildPath func(string) *Path

// Path safely holds a file path
type Path struct {
	root string // root dir
	rel  string // relative path
	sel  string // selector path
}

// NewPath returns a new Path structure based on supplied root and relative path
func newPath(root, rel string) *Path {
	return &Path{root, rel, formatSelector(rel)}
}

// NewSanitizedPathAtRoot returns a new sanitized Path structure based on root and relative path
func NewSanitizedPathAtRoot(root, rel string) *Path {
	return newPath(root, sanitizePath(rel))
}

// buildPathuserSpacesEnabled will attempt to parse a username, and return a sanitized Path at username's
// public server dir. Else, returns sanitized Path at server root
func buildPathUserSpacesEnabled(rawPath string) *Path {
	if strings.HasPrefix(rawPath, "/~") {
		// Get username and raw path
		username, rawPath := SplitByBefore(rawPath[2:], "/")

		// Treat username as a raw path, sanitizing to check for
		// dir traversals
		username = sanitizePath(username)

		// Return sanitized path using user home dir as root
		return NewSanitizedPathAtRoot("/home/"+username+"/public_"+protocol, rawPath)
	}

	// Return sanitized path at server root
	return NewSanitizedPathAtRoot(Root, rawPath)
}

// buildPathUserSpacesDisabled always returns a sanitized Path at server root
func buildPathUserSpacesDisabled(rawPath string) *Path {
	return NewSanitizedPathAtRoot(Root, rawPath)
}

// Remap remaps a Path to a new relative path, keeping previous selector
func (p *Path) Remap(newRel string) {
	p.rel = sanitizePath(newRel)
}

// RemapDirect remaps a Path to a new absolute path, keeping previous selector
func (p *Path) RemapDirect(newAbs string) {
	p.root = newAbs
	p.rel = ""
}

// Root returns file's root directory
func (p *Path) Root() string {
	return p.root
}

// Relative returns the relative path
func (p *Path) Relative() string {
	return p.rel
}

// Absolute returns the absolute path
func (p *Path) Absolute() string {
	return joinSanitizedPaths(p.root, p.rel)
}

// Selector returns the formatted selector path
func (p *Path) Selector() string {
	return p.sel
}

// RelativeDir returns the residing dir of the relative path
func (p *Path) RelativeDir() string {
	return path.Dir(p.rel)
}

// SelectorDir returns the residing dir of the selector path
func (p *Path) SelectorDir() string {
	return path.Dir(p.sel)
}

// Dir returns a Path object at the residing dir of the calling object (keeping separate selector intact)
func (p *Path) Dir() *Path {
	return &Path{p.root, p.RelativeDir(), p.SelectorDir()}
}

// JoinRelative returns a string appended to the current relative path
func (p *Path) JoinRelative(newRel string) string {
	return joinSanitizedPaths(p.rel, sanitizePath(newRel))
}

// JoinPathUnsafe appends the supplied string to the Path's relative and selector paths. This is unsafe because
// if the toJoin is a back-traversal then you can escape root
func (p *Path) JoinPathUnsafe(toJoin string) *Path {
	toJoin = sanitizePath(toJoin)
	return &Path{p.root, joinSanitizedPaths(p.rel, toJoin), joinSanitizedPaths(p.sel, toJoin)}
}

// formatSelector formats a relative path to a valid selector path
func formatSelector(rel string) string {
	if len(rel) > 0 && rel[0] == '/' {
		return rel
	}
	return "/" + rel
}

// joinSanitizedPaths quickly joins two sanitized paths, whether absolute or relative
func joinSanitizedPaths(start, end string) string {
	// Format the end string
	endLen := len(end)
	switch endLen {
	case 0:
		// do nothing
	case 1:
		// If this is '/', trim!
		if end[0] == '/' {
			end = ""
		}
	default:
		// Trim any leading '/'
		if end[0] == '/' {
			end = end[1:]
		}
		endLen--

		// Trim any trailing '/'
		if end[endLen-1] == '/' {
			end = end[:endLen-1]
		}
	}

	// Format the start string and return
	switch start {
	case "":
		return end
	case "/":
		return start + end
	default:
		// Ensure no final '/'
		length := len(start)
		if start[length-1] == '/' {
			start = start[:length-1]
		}

		// Return joined strings
		return start + "/" + end
	}
}

// pathBuilder builds relative paths,
// taking backtracks ("..") into account
type pathBuilder struct {
	b      []byte
	starts []int
}

// newPathBuilder returns a new pathBuilder
func newPathBuilder() *pathBuilder {
	return &pathBuilder{
		b:      make([]byte, 32)[:0],
		starts: []int{},
	}
}

// split add's a path split to the buffer and
// starts tracking the start of a new path segment
func (pb *pathBuilder) split() {
	// Add a '/'
	pb.b = append(pb.b, '/')

	// Set the next segment start
	pb.starts = append(pb.starts, len(pb.b))
}

// append adds a byte to the buffer
func (pb *pathBuilder) append(b byte) {
	pb.b = append(pb.b, b)
}

// backtrack removes the latest path element
func (pb *pathBuilder) backtrack() {
	// Get segments count
	length := len(pb.starts)

	// If length = 0, just empty buffer
	if length == 0 {
		pb.b = pb.b[:0]
		return
	}

	// Get the last segment start
	last := pb.starts[length-1]

	// Jump back to last segment
	pb.b = pb.b[:last-1]

	// Reset to previous segment start
	pb.starts = pb.starts[:length-1]
}

// toString gets current path as string, and resets the builder
func (pb *pathBuilder) toString() string {
	// Get buffer as string, skip first
	// char i.e. '/'
	s := string(pb.b)
	if len(s) > 0 {
		s = s[1:]
	}

	// Reset buffer and list of segment starts
	pb.b = pb.b[:0]
	pb.starts = pb.starts[:0]

	// Return string
	return s
}

// sanitizePath sanitizes a raw path
func sanitizePath(raw string) string {
	// Empty path, nothing to do
	if raw == "" {
		return ""
	}

	// Get necessary information beforehand
	length := len(raw)
	pb := pathBuilderPool.Get().(*pathBuilder)
	defer pathBuilderPool.Put(pb)

	for index := 0; index < length; {
		// Path segment separator
		if raw[index] == '/' {
			index++
			continue
		}

		// Segment starting '.'
		if raw[index] == '.' {
			// Hit the end of the path, break-out
			if index+1 == length {
				break
			}

			// Iter index
			index++

			// This segment is only '.', continue
			if raw[index] == '/' {
				index++
				continue
			}

			// If next char is '.'
			if raw[index] == '.' {
				// Hit end of the path, break-out
				if index+1 == length {
					pb.backtrack()
					break
				}

				// Iter index
				index++

				// If next char is '/' then this is a
				// back-track segment.
				if raw[index] == '/' {
					pb.backtrack()
					continue
				}

				// Split for next segment
				pb.split()

				// Else, add a '.'
				pb.append('.')
			} else {
				// Split for next segment
				pb.split()
			}

			// Else, append a '.'
			pb.append('.')

			// Continue adding up to next '/'
			for ; index < length && raw[index] != '/'; index++ {
				pb.append(raw[index])
			}

			// Final iter to skip past the final '/'
			index++
			continue
		}

		// Split for next segment
		pb.split()

		// Iter through and add up to next '/'
		for ; index < length && raw[index] != '/'; index++ {
			pb.append(raw[index])
		}

		// Final iter to skip past the final '/'
		index++
	}

	// Return built string
	return pb.toString()
}
